@@ -12,3 +12,7 @@
 
 #define QT_FEATURE_networklistmanager -1
 
+#define QT_FEATURE_publicsuffix_qt 1
+
+#define QT_FEATURE_publicsuffix_system 1
+
